/*
    UPN-Calc - An arbitrary precision RPN calculator for J2ME/MIDP1
    Copyright (C) 2005 Kai G. Schwebke, upn.schwebke.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/*
 * Arbitrary Precision Unnormalized Floating Point Arithmetic
 */

class BCDFloat
{
   BCDInteger c; // coefficient
   int e;        // exponent
   int prec;     // precision

   // Special values
   int type;
   static final int FINITE = 0;
   static final int NAN = 1;
   static final int P_INF = 2;
   static final int N_INF = 3;

   static final int MAX_E = 1000;


   private static int defaultPrec = 14;

   public static void setDefaultPrec(int p)
   {
      defaultPrec = p;
      BCDConst.reset();
   }

   public static int getDefaultPrec()
   {
      return defaultPrec;
   }

   public BCDFloat()
   {
      init();
   }

   public BCDFloat(String s)
   {
      init();
      init(s);
   }

   public BCDFloat(int i)
   {
      init();
      c.initFromInteger(i);
   }

   public void init()
   {
      prec = defaultPrec;
      e = 0;
      c = new BCDInteger();
      type = FINITE;
   }

   public void init(String s)
   {
      if (s.equals("NaN"))
      {
	 type = NAN;
	 return;
      }
      if (s.equals("+Inf"))
      {
	 type = P_INF;
	 return;
      }
      if (s.equals("-Inf"))
      {
	 type = N_INF;
	 return;
      }

      // search for exponent, parse, cut
      int idx = s.indexOf('e');
      if (idx == (-1))
      {
	 idx = s.indexOf('E');
      }
      if (idx != (-1))
      {
	 String es = s.substring(idx+1);
	 try {
	    e = Integer.parseInt(es);
	 } catch (NumberFormatException e) {
	 }
	 s = s.substring(0, idx);
      }

      // search for decimal point, cut
      idx = s.indexOf('.');
      if (idx != (-1))
      {
	 e -= s.length()-idx-1;
	 s = s.substring(0, idx) + s.substring(idx+1);
      }

      c.initFromString(s);
   }

   public BCDFloat copy()
   {
      BCDFloat r = new BCDFloat();
      r.c = c.copy();
      r.e = e;
      r.type = type;
      return r;
   }

   public void normalize()
   {
      if (type != FINITE)
      {
	 return;
      }

      c.shorten();
      int l = c.numDigits;

      while ( (l > 1) && (c.d[0] == 0) )
      {
	 c = c.shiftLeft(1);
	 ++e;
	 --l;
      }
      if (l == 0)
      {
	 e = 0;
      }
   }

   public void normalize(int maxE)
   {
      if (type != FINITE)
      {
	 return;
      }

      c.shorten();
      int l = c.numDigits;

      while ( (l > 1) && (c.d[0] == 0) && (e < maxE) )
      {
	 c = c.shiftLeft(1);
	 ++e;
	 --l;
      }
      if (l == 0)
      {
	 e = 0;
      }
   }

   public void round()
   {
      if (type != FINITE)
      {
	 return;
      }

      c.shorten();
      int l = c.numDigits;

      if (l > prec)
      {
	 boolean roundUp = false;
	 if (c.d[l-prec-1] >= 5)
	 {
	    roundUp = true;
	 }

	 c = c.shiftLeft(l-prec);
	 e += l-prec;
	 if (roundUp)
	 {
	    c = c.addUnsigned(1);
	 }
      }

      if (e > MAX_E)
      {
	 if (c.neg)
	 {
	    type = N_INF;
	 } else {
	    type = P_INF;
	 }
      }
      if (e < -MAX_E)
      {
	 c = new BCDInteger();
	 e = 0;
      }
   }

   public boolean isZero()
   {
      return ((type == FINITE) && (c.numDigits == 0));
   }

   public BCDFloat add(BCDFloat v)
   {
      BCDFloat r = new BCDFloat();

      if ((type == NAN) ||
	  (v.type == NAN) ||
          ((type == P_INF)&&(v.type == N_INF)) ||
	   ((type == N_INF)&&(v.type == P_INF)))
      {
	 r.type = NAN;
	 return r;
      }
      if ((type == P_INF) || (v.type == P_INF))
      {
	 r.type = P_INF;
	 return r;
      }
      if ((type == N_INF) || (v.type == N_INF))
      {
	 r.type = N_INF;
	 return r;
      }

      int de = Math.abs(e - v.e);
      if (de >= 2*prec+2)
      {
	 if (e > v.e)
	 {
	    r.c = c.copy();
	    r.e = e;
	 } else {
	    r.c = v.c.copy();
	    r.e = v.e;
	 }
      } else {
	 BCDInteger coeff = c;
	 BCDInteger coeffV = v.c;
	 r.e = v.e;
	 if (e > v.e)
	 {
	    coeff = c.shiftRight(de);
	 }
	 if (v.e > e)
	 {
	    coeffV = v.c.shiftRight(de);
	    r.e = e;
	 }
	 r.c = coeff.add(coeffV);
	 r.round();
      }

      return r;
   }

   public BCDFloat sub(BCDFloat vt)
   {
      BCDFloat v = vt.copy();
      v.c.neg = !v.c.neg;
      if (v.type == P_INF)
      {
	 v.type = N_INF;
      } else if (v.type == N_INF) {
	 v.type = P_INF;
      }
      BCDFloat r = add(v);
      return r;
   }

   public BCDFloat mul(BCDFloat v)
   {
      BCDFloat p = new BCDFloat();

      if ((type == NAN) ||
	  (v.type == NAN))
      {
	 p.type = NAN;
	 return p;
      }
      if (((type == P_INF)&&(v.type == P_INF)) ||
	  ((type == N_INF)&&(v.type == N_INF)))
      {
	 p.type = P_INF;
	 return p;
      }
      if (((type == P_INF)&&(v.type == N_INF)) ||
	  ((type == N_INF)&&(v.type == P_INF)))
      {
	 p.type = N_INF;
	 return p;
      }
      if (((type == N_INF)&&(!v.c.neg)) ||
	  ((type == P_INF)&&(v.c.neg)) ||
	  ((v.type == N_INF)&&(!c.neg)) ||
	  ((v.type == P_INF)&&(c.neg)))
      {
	 p.type = N_INF;
	 return p;
      }
      if (((type == P_INF)&&(!v.c.neg)) ||
	  ((type == N_INF)&&(v.c.neg)) ||
	  ((v.type == P_INF)&&(!c.neg)) ||
	  ((v.type == N_INF)&&(c.neg)))
      {
	 p.type = N_INF;
	 return p;
      }

      p.c = c.mul(v.c);
      p.e = e + v.e;

      p.round();

      return p;
   }

   public BCDFloat div(BCDFloat v)
   {
      BCDFloat q = new BCDFloat();

      if ((type == NAN) ||
	  (v.type == NAN) ||
	  (isZero() && v.isZero()) ||
	  (((type == P_INF)||(type == N_INF))&&
	   ((v.type == P_INF)||(v.type == N_INF))))
      {
	 q.type = NAN;
	 return q;
      }
      if (((type == P_INF)&&(!v.c.neg)) ||
	  ((type == N_INF)&&(v.c.neg)))
      {
	 q.type = P_INF;
	 return q;
      }
      if (((type == N_INF)&&(!v.c.neg)) ||
	  ((type == P_INF)&&(v.c.neg)))
      {
	 q.type = N_INF;
	 return q;
      }
      if ((v.type == P_INF)||(v.type == N_INF))
      {
	 return q;
      }

      if (v.isZero())
      {
	 if (c.neg)
	 {
	    q.type = N_INF;
	 } else {
	    q.type = P_INF;
	 }
	 return q;
      }

      
      int idealExp = e-v.e;
      q.e = idealExp;

      BCDInteger coeffU = c.copy();
      coeffU.shorten();
      BCDInteger coeffV = v.c.copy();
      coeffV.shorten();

      if (coeffU.numDigits < prec+1+coeffV.numDigits)
      {
	 int corr = prec+1+coeffV.numDigits-coeffU.numDigits;
	 coeffU = coeffU.shiftRight(corr);
	 q.e -= corr;
      }

      q.c = coeffU.div(coeffV);
      q.round();

      if (q.e < idealExp)
      {
	 q.normalize(idealExp);
      }
      if (q.e > idealExp)
      {
	 q.c = q.c.shiftRight(q.e-idealExp);
	 q.e -= q.e-idealExp;
      }

      return q;
   }

   public BCDFloat sqrt()
   {
      // start value: input value with half of effective exponent
      BCDFloat r = new BCDFloat();

      if (c.neg)
      {
	 r.type = NAN;
	 return r;
      }

      r.c = c.copy();
      r.e = (e-c.numDigits)/2;

      BCDFloat oneHalf = BCDConst.getHalf();

      BCDFloat rn;
      // Newton iteration
      int i;
      for (i = 0; i < 10*prec; ++i) // max. number of iterations
      {
	 rn = r.add(this.div(r)).mul(oneHalf);
	 if ( rn.sub(r).isZero() ||
	      (rn.type != FINITE) ) // stop if increment is zero
	 {
	    r = rn;
	    break;
	 }
	 r = rn;
      }

      r.normalize((e+1)/2);

      return r;
   }
      
   public BCDFloat exp() // e^x based on expI using transformation
   {                     // to interval [0, ln 10] using:
			 // x - k * ln 10 = r
			 // 10^k * e^r = e^x
      
      BCDFloat ln10 = BCDConst.getLn10();
      int k = this.div(ln10).toInteger();

      if (k == 0)
      {
	 return expI();
      }

      BCDFloat r = this.sub((new BCDFloat(k)).mul(ln10));

      r = r.expI();

      r.e += k;

      return r;
   }

   public BCDFloat expI() // e^x using series expansion, usable around 0
   {
      BCDFloat rn;
      BCDFloat a = this.copy();
      BCDFloat b = BCDConst.getOne();
      BCDFloat r = b.add(this); // 1+x

      // Taylor Series Expansion
      int i;
      for (i = 2; i < 10*prec; ++i) // max. number of iterations
      {
	 a = a.mul(this);
	 b = b.mul(new BCDFloat(i));
	 rn = r.add(a.div(b));
	 
	 if ( rn.sub(r).isZero() ||
	      (rn.type != FINITE) ) // stop if increment is zero
	 {
	    r = rn;
	    break;
	 }

	 r = rn;
      }

      return r;
   }

   public BCDFloat ln()
   {
      if (isZero())
      {
	 BCDFloat r = new BCDFloat();
	 r.type = N_INF;
	 return r;
      }
      if (c.neg)
      {
	 BCDFloat r = new BCDFloat();
	 r.type = NAN;
	 return r;
      }

      BCDFloat one = BCDConst.getOne();
      BCDFloat e = BCDConst.getE();
      BCDFloat eInv = BCDConst.getEInv();
      BCDFloat x = this;

      boolean cInv = false;
      if (!x.sub(one).c.neg) // x > 1 use x' = 1/x, ln x' = - ln x
      {
	 cInv = true;
	 x = one.div(x);
      }

      // determine k (min, integer) such that e^k*x = r > 1/e
      BCDFloat r = x;
      BCDFloat ep = e;

      //    calculate max n (step) required with k = n
      int p = 1;
      while (r.sub(eInv).c.neg)
      {
	 p *= 2;
	 ep = ep.mul(ep);

	 r = x.mul(ep);
      }

      //    using max step approximate k with bisection
      //    (decreasing number of multiplications for large k)
      int k = 0;
      r = x;
      for (p/=2; p >= 1; p/=2)
      {
	 // prepare step
	 int pt = 1;
	 ep = e;
	 while (pt != p)
	 {
	    pt *= 2;
	    ep = ep.mul(ep);
	 }

	 int kp = k;
	 BCDFloat rp = r;

	 // add step, rollback on overflow
	 while (r.sub(eInv).c.neg)
	 {
	    kp = k;
	    rp = r;

	    r = r.mul(ep);
	    k += p;
	 }
	 r = rp;
	 k = kp;
      }

      // ln x = ln (r * e^(-k) ) = ln r - k
      r = r.lnI().sub(new BCDFloat(k));

      if (cInv) // x > 1
      {
	 r.c.neg = !r.c.neg;
      }

      return r;
   }

   public BCDFloat lnI() // ln x using series expansion, usable around 1
   {
      BCDFloat b;
      BCDFloat d = this.sub(BCDConst.getOne()); // 1-x
      BCDFloat a = d.copy();
      BCDFloat r = d.copy();
      BCDFloat rn;

      // Taylor Series Expansion
      int i;
      for (i = 2; i < 10*prec; ++i) // max. number of iterations
      {
	 a = a.mul(d);
	 b = new BCDFloat(i);
	 if (i%2 == 0)
	 {
	    rn = r.sub(a.div(b));
	 } else {
	    rn = r.add(a.div(b));
	 }
	 
	 if ( rn.sub(r).isZero() ||
	      (rn.type != FINITE) ) // stop if increment is zero
	 {
	    r = rn;
	    break;
	 }

	 r = rn;
      }

      return r;
   }

   public BCDFloat cos()
   {
      return this.add(BCDConst.getPiHalf()).sin();
   }

   public BCDFloat tan()
   {
      return this.sin().div(this.cos());
   }

   public BCDFloat sin()
   {
      BCDFloat x = this.copy();
      boolean xNeg = false;

      // sin x = - sin(-x)
      if (x.c.neg)
      {
	 xNeg = true;
	 x.c.neg = false;
      }

      if (!x.sub(BCDConst.get2Pi()).c.neg) // x > 2 pi
      {
	 int k = x.div(BCDConst.get2Pi()).toInteger();
	 x = x.sub(BCDConst.get2Pi().mul(new BCDFloat(k)));
      }

      if (!x.sub(BCDConst.getPi()).c.neg) // Pi < x < 2 Pi
      {
	 xNeg = !xNeg;
	 x = x.sub(BCDConst.getPi());
      }

      BCDFloat s = x.sinI();

      if (xNeg)
      {
	 s.c.neg = !s.c.neg;
      }

      return s;
   }

   public BCDFloat sinI() // sin using series expansion,
   {                      // usable around 0 (-Pi..Pi)
      BCDFloat x = this;
      BCDFloat xsq = x.mul(x);
      BCDFloat f = BCDConst.getOne();

      BCDFloat s = x;
      BCDFloat sn;

      int i;
      for (i = 1; i < 10*prec; ++i)
      {
	 x = x.mul(xsq);
	 f = f.mul(new BCDFloat(2*i)).mul(new BCDFloat(2*i+1));

	 if (i%2 != 0)
	 {
	    sn = s.sub(x.div(f));
	 } else {
	    sn = s.add(x.div(f));
	 }

	 if ( sn.sub(s).isZero() ||
	      (sn.type != FINITE) ) // stop if increment is zero
	 {
	    s = sn;
	    break;
	 }

	 s = sn;
      }

      return s;
   }

   public BCDFloat atan()
   {
      BCDFloat x = this.copy();
      boolean xNeg = false;

      if (x.c.neg)
      {
	 x.c.neg = false;
	 xNeg = true;
      }


      boolean gtOne = false;
      if (!x.sub(BCDConst.getOne()).c.neg) // x > 1
      {
	 gtOne = true;
	 x = BCDConst.getOne().div(x);
      }


      BCDFloat at = x;
      BCDFloat atn;

      int i;
      for (i = 1; i < prec; ++i)
      {
	 BCDFloat atTan = at.tan();
	 atn = at.sub(atTan.sub(x).div(BCDConst.getOne().add(atTan.mul(atTan))));

	 if ( atn.sub(at).isZero() ||
	      (atn.type != FINITE) ) // stop if increment is zero
	 {
	    atn = at;
	    break;
	 }

	 at = atn;
      }



      if (gtOne)
      {
	 at = BCDConst.getPiHalf().sub(at);
      }

      if (xNeg)
      {
	 at.c.neg = !at.c.neg;
      }

      return at;
   }

   public BCDFloat asin()
   {
      BCDFloat x = this.copy();
      boolean xNeg = false;

      if (x.c.neg)
      {
	 x.c.neg = false;
	 xNeg = true;
      }

      
      BCDFloat xTest = x.sub(BCDConst.getOne());
      BCDFloat as = BCDConst.getPiHalf();

      if (!xTest.isZero())
      {
	 if (!xTest.c.neg) // x > 1
	 {
	    as = new BCDFloat();
	    as.type = NAN;
	 } else {
	    as = x.div(BCDConst.getOne().sub(x.mul(x)).sqrt()).atan();
	 }
      }

      if (xNeg)
      {
	 as.c.neg = !as.c.neg;
      }

      return as;
   }

   public BCDFloat acos()
   {
      BCDFloat x = this.copy();
      x.c.neg = !x.c.neg;
      BCDFloat ac = x.asin();
      ac = ac.add(BCDConst.getPiHalf());

      return ac;
   }

   // convert to native/debug string representation
   public String toNativeString()
   {
      String s = "";

      if (type == FINITE)
      {
	 s = "(";
	 s += c.toString();
	 s += ","; 
	 s += e;
	 s += ")";
      }

      if (type == NAN)
      {
	 s = "(NaN)";
      }

      if (type == P_INF)
      {
	 s = "(+Inf)";
      }

      if (type == N_INF)
      {
	 s = "(-Inf)";
      }

      return s;
   }

   // convert to scientific string representation
   public String toString()
   {
      String r;

      if (type == NAN)
      {
	 r = "NaN";
	 return r;
      }

      if (type == P_INF)
      {
	 r = "+Inf";
	 return r;
      }

      if (type == N_INF)
      {
	 r = "-Inf";
	 return r;
      }

      String s = c.toString();
      String sign = "";
      if (c.neg)
      {
	 s = s.substring(1);
	 sign = "-";
      }

      int adjExp = e + s.length() - 1;

      if ( (e <= 0) && (adjExp >= (-6)) ) // no E-notation
      {
	 if (e == 0) // no decimal point
	 {
	    r = sign + s;
	 } else { // decimal point
	    int l = s.length();
	    while (l < (-e + 1))
	    {
	       s = "0"+s;
	       ++l;
	    }
	    r = sign + s.substring(0, l+e) + "." + s.substring(l+e);
	 }
      } else { // E-notation
	 int l = s.length();
	 if (l > 1) // decimal point
	 {
	    s = s.substring(0, 1) + "." + s.substring(1);
	 }
	 r = sign + s + "E" + adjExp;
      }

      return r;
   }

   public int toInteger()
   {
      if (e > 0)
      {
	 c = c.shiftRight(e);
	 e = 0;
      }
      if (e < 0)
      {
	 if (c.numDigits > (-e))
	 {
	    c = c.shiftLeft(-e);
	    e = 0;
	 } else {
	    return 0;
	 }
      }
      return c.toInteger();
   }

}
